# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 07.11.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 11, task 6.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation
# pylint: disable=redefined-outer-name
# pylint: disable=import-error

import os

import numpy as np
import matplotlib.pyplot as plt
import cv2
import skimage.morphology as morph
import skimage.segmentation as segmentation
from skimage.feature import peak_local_max
from skimage.color import label2rgb
#import scipy.ndimage as ndi

import skiz

def plot_image(image, fig_num, name=None, colormap='gray', write_file=None,
               fixed_range=True):
  """Plot image"""
  fig = plt.figure(fig_num)
  if fixed_range:
    vmin = np.iinfo(image.dtype).min
    vmax = np.iinfo(image.dtype).max
  else:
    vmin = int(np.min(image))
    vmax = int(np.max(image))
  plt.imshow(image, cmap=colormap, interpolation='none', vmin=vmin, vmax=vmax)
  if name is not None:
    plt.title(name)
  plt.xticks([]), plt.yticks([])
  plt.tight_layout()
  if write_file:
    fig.savefig(write_file, bbox_inches='tight', pad_inches=0)
  fig_num += 1
  return fig_num


def compute_gradient_magnitude(image):
  """Compute gradient magnitude of input image"""

  # Use Sobel filter to approximate the horizontal and vertical gradient
  sobelx = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
  sobely = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)

  # Compute gradient magnitude
  grad_magn = np.sqrt(sobelx**2 + sobely**2)

  # Put it in [0, 255] value range
  grad_magn = 255*(grad_magn - np.min(grad_magn)) / (np.max(grad_magn) - np.min(grad_magn))

  return grad_magn.astype(np.uint8)

def imregionalmax(image, ksize=3):
  """Similar to matlab's imregionalmax"""
  filterkernel = np.ones((ksize, ksize)) # 8-connectivity
  reg_max_loc = peak_local_max(image,
                               footprint=filterkernel, indices=False,
                               exclude_border=0)
  return reg_max_loc.astype(np.uint8)

def imcomplement(image):
  """Equivalent to matlabs imcomplement function"""
  min_type_val = np.iinfo(image.dtype).min
  max_type_val = np.iinfo(image.dtype).max
  return min_type_val + max_type_val - image


def main():
  """Binary segmentation"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 11')
  print('Task 6')
  print('-'*80)

  # Configurations
  result_dir = 'results/images'
  image_file = '../../images/pears.png'
  use_skiz = False

  # Step 1: Read image
  image = cv2.imread(image_file, cv2.IMREAD_GRAYSCALE)
  fig_num = plot_image(image, 0,
                       write_file=os.path.join(result_dir, 't06_gray_pears.png'))


  # Step 2: Use gradient magnitude as segmentation function (bad idea in this case)
  grad_magn = compute_gradient_magnitude(image)
  fig_num = plot_image(grad_magn, fig_num,
                       write_file=os.path.join(result_dir, 't06_grad_magn.png'))


  # Try to segment using the complement of the thresholded gradient magnitude
  # as seed points (bad idea)
  #_, thresh_magn = cv2.threshold(grad_magn, 0, 255,
  #                               cv2.THRESH_BINARY+cv2.THRESH_OTSU)
  #_, labeled_thresh = cv2.connectedComponents(255 - thresh_magn)
  #grad_ws = morph.watershed(grad_magn, labeled_thresh)
  #col_ws = label2rgb(grad_ws)
  #fig_num = plot_image(col_ws, fig_num, fixed_range=False)

  # Step 3: Mark foreground objects

  ## We will use morphological methods to "clean" the image.

  selem = morph.disk(20)

  ### Opening
  ### NB: If you are following the matlab solution proposal, you will notice
  ### slight differences (probably due to different connectivity or structure
  ### elements)
  opened = morph.opening(image, selem)
  fig_num = plot_image(opened, fig_num,
                       write_file=os.path.join(result_dir, 't06_opened.png'))

  ### Reconstruction by dilation (opening by reconstruction)
  eroded = morph.erosion(image, selem)
  opening_recon = morph.reconstruction(eroded, image, method='dilation').astype(np.uint8)
  fig_num = plot_image(opening_recon, fig_num,
                       write_file=os.path.join(result_dir, 't06_opening_recon.png'))

  ### Opening followed by closing
  closed_opening = morph.closing(opened, selem)
  fig_num = plot_image(closed_opening, fig_num,
                       write_file=os.path.join(result_dir, 't06_closed_opening.png'))

  ### Closing followed by opening
  #closed = morph.closing(image, selem)
  #fig_num = plot_image(closed, fig_num,
  #                     write_file=os.path.join(result_dir, 't06_closed.png'))
  #opened_closing = morph.opening(closed, selem)
  #fig_num = plot_image(opened_closing, fig_num,
  #                     write_file=os.path.join(result_dir, 't06_opened_closing.png'))

  ### Reconstruction by erosion (closing by reconstruction)
  #dilated = morph.dilation(image, selem)
  #closing_recon = morph.reconstruction(dilated, image, method='erosion')
  #fig_num = plot_image(closing_recon, fig_num,
  #                     write_file=os.path.join(result_dir, 't06_closing_recon.png'))

  # Use reconstruction by erosion on the image you got from reconstruction by
  # dilation
  dilated_recon_dilation = morph.dilation(opening_recon, selem)
  recon_erosion_recon_dilation = morph.reconstruction(dilated_recon_dilation,
                                                      opening_recon,
                                                      method='erosion').astype(np.uint8)
  fig_num = plot_image(recon_erosion_recon_dilation, fig_num,
                       write_file=os.path.join(result_dir, 't06_recon_erosion_dilation.png'))

  # Note: In the matlab solution proposal, they complement the input in the
  # opening by reconstruction and then complement the output. This is
  # equivalent to the closing-by-reconstruction which is done above, uncomment
  # the below code to check.
  #recon_dilation_recon_dilation = morph.reconstruction(imcomplement(dilated_recon_dilation),
  #                                                     imcomplement(opening_recon),
  #                                                     method='dilation').astype(np.uint8)
  #recon_dilation_recon_dilation_c = imcomplement(recon_dilation_recon_dilation)
  #print(np.linalg.norm(recon_erosion_recon_dilation - recon_dilation_recon_dilation_c))
  ## Returns 0.0
  #print(np.allclose(recon_erosion_recon_dilation, recon_dilation_recon_dilation_c))
  ## Returns True

  # Get regional max to be used as sure foreground
  #reg_max = ndi.filters.maximum_filter(recon_erosion_recon_dilation)
  #reg_max_mask = (reg_max == recon_erosion_recon_dilation) # Convert local max
                                                            # to binary mask

  foreground_1 = imregionalmax(recon_erosion_recon_dilation, ksize=65)
  fig_num = plot_image(foreground_1*255, fig_num,
                       write_file=os.path.join(result_dir, 't06_foreground_1.png'))

  fg_superimposed_1 = image.copy()
  fg_superimposed_1[foreground_1 == 1] = 255
  fig_num = plot_image(fg_superimposed_1, fig_num,
                       write_file=os.path.join(result_dir, 't06_foreground_1_superimposed.png'))

  # Clean foreground
  foreground_2 = morph.closing(foreground_1, np.ones((5, 5)))
  foreground_3 = morph.erosion(foreground_2, np.ones((5, 5)))
  #fig_num = plot_image(foreground_3*255, fig_num)


  # Remove small objects
  foreground_4 = morph.remove_small_objects(foreground_3.astype(bool), min_size=20)
  #fig_num = plot_image(foreground_4*255, fig_num)

  # Superimpose the foreground on the original image
  _, labeled_fg = cv2.connectedComponents(foreground_4.astype(np.uint8))
  col_labeled_fg = label2rgb(labeled_fg, image)
  fig_num = plot_image(col_labeled_fg, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't06_foreground_4_superimposed.png'))


  ### Step 4 Compute background markers

  _, thresholded = cv2.threshold(recon_erosion_recon_dilation, 0, 255,
                                 cv2.THRESH_BINARY+cv2.THRESH_OTSU)
  fig_num = plot_image(thresholded, fig_num,
                       write_file=os.path.join(result_dir, 't06_thresholded.png'))

  #### Optional: Compute the SKIZ, and use it to enforce sure boundaries
  if use_skiz:
    skiz_im = skiz.skiz(thresholded > 0)
    fig_num = plot_image((skiz_im*255).astype(np.uint8), fig_num,
                         write_file=os.path.join(result_dir, 't06_skiz.png'))


  ### Step 5: Compute the watershed transform of the segmentation function

  if use_skiz:
    grad_magn = grad_magn + (skiz_im*255).astype(np.uint8)


  labels = morph.watershed(grad_magn, labeled_fg)

  ## Step 6: Visualize the result

  ### Method 1: Superimpose watershed boundaries, foreground and skiz on the original image
  superimposed = image.copy()
  watershed_boundaries = segmentation.find_boundaries(labels)
  superimposed[watershed_boundaries] = 255
  superimposed[foreground_4] = 255
  if use_skiz:
    superimposed[skiz_im] = 255

  fig_num = plot_image(superimposed, fig_num,
                       write_file=os.path.join(result_dir, 't06_superimposed.png'))

  #### Method2: Superimpose the watershed labels on the original image

  col_labels = label2rgb(labels)
  fig_num = plot_image(col_labels, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't06_ws_labels.png'))

  col_labels_merged = label2rgb(labels, image)
  fig_num = plot_image(col_labels_merged, fig_num, fixed_range=False,
                       write_file=os.path.join(result_dir, 't06_ws_label_overlayed.png'))



  plt.show()

if __name__ == "__main__":
  main()
